@@ -1,4 +1,5 @@
-/* $Id: autowhite.h,v 1.16 2004/09/13 18:41:54 manu Exp $ */
+/* $Id: autowhite.h,v 1.21 2006/08/30 04:57:58 manu Exp $ */
+/* vim: set sw=8 ts=8 sts=8 noet cino=(0: */
 
 /*
  * Copyright (c) 2004 Emmanuel Dreyfus
@@ -38,9 +39,9 @@
 #define AUTOWHITE_VALIDITY (24 * 3600) /* 1 day */
 #endif
 
-#define AUTOWHITE_WRLOCK WRLOCK(autowhite_lock) 
-#define AUTOWHITE_RDLOCK RDLOCK(autowhite_lock) 
-#define AUTOWHITE_UNLOCK UNLOCK(autowhite_lock)
+#ifndef AUTOWHITE_BUCKETS
+#define AUTOWHITE_BUCKETS 512
+#endif
 
 TAILQ_HEAD(autowhitelist, autowhite);
 
@@ -51,17 +52,26 @@ struct autowhite {
 	char *a_rcpt;
 	struct timeval a_tv;
 	TAILQ_ENTRY(autowhite) a_list;
+	TAILQ_ENTRY(autowhite) ab_list;
 };
 
-extern pthread_rwlock_t autowhite_lock;
+struct autowhite_bucket {
+	TAILQ_HEAD(, autowhite) b_autowhite_head;
+};
+
+#define AUTOWHITE_LOCK pthread_mutex_lock(&autowhite_lock)
+#define AUTOWHITE_UNLOCK pthread_mutex_unlock(&autowhite_lock)
+extern pthread_mutex_t autowhite_lock;
 
 void autowhite_init(void);
 struct autowhite *autowhite_get(struct sockaddr *, socklen_t, char *, char *,
-    time_t *);
+    time_t);
 void autowhite_put(struct autowhite *);
 void autowhite_add(struct sockaddr *, socklen_t, char *, char *, time_t *,
     char *);
-int autowhite_check(struct sockaddr *, socklen_t, char *, char *, char *);
+int autowhite_check(struct sockaddr *, socklen_t, char *, char *, 
+    char *, time_t, time_t);
 int autowhite_textdump(FILE *);
+int autowhite_del_addr(struct sockaddr *, socklen_t);
 
 #endif /* _AUTOWHITE_H_ */
