@@ -1,4 +1,4 @@
-/* $Id: conf.h,v 1.31 2005/03/19 07:38:53 manu Exp $ */
+/* $Id: conf.h,v 1.40 2007/02/06 14:29:55 manu Exp $ */
 
 /*
  * Copyright (c) 2004 Emmanuel Dreyfus
@@ -42,7 +42,6 @@
 #include <stdio.h>
 #include <pthread.h>
 #include <sys/socket.h>
-#include <sys/time.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
@@ -52,12 +51,16 @@
 #define CONFFILE "/etc/mail/greylist.conf"
 #endif
 
-#define CONF_WRLOCK WRLOCK(conf_lock) 
-#define CONF_RDLOCK RDLOCK(conf_lock) 
-#define CONF_UNLOCK UNLOCK(conf_lock)
+#ifndef DRACDB
+#define DRACDB "/usr/local/etc/dracdb.db"
+#endif
+
+TAILQ_HEAD(conf_list, conf_rec);
+struct conf_rec {
+	int c_refcount;
+	time_t c_timestamp;
+	TAILQ_ENTRY(conf_rec) c_chain;
 
-struct conf {
-	int c_cold;
 	int c_forced;
 	int c_debug;
 	int c_acldebug;
@@ -76,15 +79,32 @@ struct conf {
 	struct in6_addr c_match_mask6;
 #endif
 	char *c_socket;
+	int c_socket_mode;
 	char *c_user;
 	char *c_syncaddr;
 	char *c_syncport;
+	char *c_syncsrcaddr;
+	char *c_syncsrcport;
 	int c_nodetach;
 	int c_report;
 	int c_lazyaw;
 	int c_dumpfreq;
 	int c_timeout;
 	int c_extendedregex;
+	char *c_dracdb;
+	int c_nodrac;
+	int c_dump_no_time_translation;
+	int c_logexpired;
+	char c_pidfile_storage[QSTRLEN + 1];
+	char c_dumpfile_storage[QSTRLEN + 1];
+	char c_socket_storage[QSTRLEN + 1];
+	char c_user_storage[QSTRLEN + 1];
+	char c_syncaddr_storage[IPADDRSTRLEN + 1];
+	char c_syncport_storage[NUMLEN + 1];
+	char c_syncsrcaddr_storage[IPADDRSTRLEN + 1];
+	char c_syncsrcport_storage[NUMLEN + 1];
+	char c_dracdb_storage[QSTRLEN + 1];
+	size_t c_maxpeek;
 };
 
 /* c_forced flags */
@@ -113,26 +133,29 @@ struct conf {
 #define C_NODELAYS	0x2
 #define C_ALL		0x3
 
-extern struct conf defconf;
-extern struct conf conf;
+extern struct conf_rec defconf;
+extern pthread_key_t conf_key;
+#define GET_CONF() ((struct conf_rec *)pthread_getspecific(conf_key))
+#define conf (*GET_CONF())
 extern char *conffile;
-extern char c_pidfile[PATHLEN + 1];
-extern char c_dumpfile[PATHLEN + 1];
-extern char c_socket[PATHLEN + 1];
-extern char c_user[PATHLEN + 1];
-extern char c_syncaddr[IPADDRSTRLEN + 1];
-extern char c_syncport[NUMLEN + 1];
-
-extern pthread_rwlock_t conf_lock;
+extern int conf_cold;
 
 void conf_init(void);
 void conf_load(void);
 void conf_update(void);
+void conf_retain(void);
+void conf_release(void);
 
 extern FILE *conf_in;
 extern int conf_line;
+extern int conf_acl_end;
+extern int conf_racl_end;
+extern int conf_dacl_end;
+
 int conf_parse(void);
+void conf_dispose_input_file(void);
 char *quotepath(char *, char *, size_t);
-void conf_defaults(struct conf *);
+char *quotepath_alloc(const char *);
+void conf_defaults(struct conf_rec *);
 
 #endif /* _CONF_H_ */
