@@ -1,4 +1,4 @@
-/* $Id: pending.c,v 1.67 2005/05/11 16:25:43 manu Exp $ */
+/* $Id: pending.c,v 1.85 2007/10/18 15:07:47 manu Exp $ */
 
 /*
  * Copyright (c) 2004 Emmanuel Dreyfus
@@ -34,7 +34,7 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID  
-__RCSID("$Id: pending.c,v 1.67 2005/05/11 16:25:43 manu Exp $");
+__RCSID("$Id: pending.c,v 1.85 2007/10/18 15:07:47 manu Exp $");
 #endif
 #endif
 
@@ -46,6 +46,7 @@ __RCSID("$Id: pending.c,v 1.67 2005/05/11 16:25:43 manu Exp $");
 
 #include <stdlib.h>
 #include <stdio.h>
+#include <ctype.h>
 #include <string.h>
 #include <strings.h>
 #include <pthread.h>
@@ -69,34 +70,69 @@ __RCSID("$Id: pending.c,v 1.67 2005/05/11 16:25:43 manu Exp $");
 #include "autowhite.h"
 #include "milter-greylist.h"
 
+#ifdef USE_DMALLOC
+#include <dmalloc.h> 
+#endif
+
+struct pending_bucket *pending_buckets;
 struct pendinglist pending_head;
-pthread_rwlock_t pending_lock; 	/* protects pending_head and dump_dirty */
+/* protects pending_head and pending_buckets */
+pthread_mutex_t pending_lock = PTHREAD_MUTEX_INITIALIZER;
+
 /*
  * protects pending->p_refcnt
  * since we hold many pending entries, requied memory for the lock
  * object is considerably large to have it in each pending entry.  so,
  * we use just one lock object for all pending entries.
  */
-static pthread_rwlock_t refcnt_lock;
+static pthread_mutex_t refcnt_lock = PTHREAD_MUTEX_INITIALIZER;
 
 void
 pending_init(void) {
-	int error;
+	int i;
 
 	TAILQ_INIT(&pending_head);
-	if ((error = pthread_rwlock_init(&pending_lock, NULL)) != 0 ||
-	    (error = pthread_rwlock_init(&refcnt_lock, NULL)) != 0) {
-		syslog(LOG_ERR, 
-		    "pthread_rwlock_init failed: %s", strerror(error));
+	if ((pending_buckets = calloc(PENDING_BUCKETS, 
+	    sizeof(struct pending_bucket))) == NULL) {
+		mg_log(LOG_ERR, 
+		    "Unable to allocate pending list buckets: %s", 
+		    strerror(errno));
 		exit(EX_OSERR);
 	}
+	
+	for(i = 0; i < PENDING_BUCKETS; i++) {
+		TAILQ_INIT(&pending_buckets[i].b_pending_head);
+	}
 
 	return;
 }
 
+/* pending_lock must be locked */
+int 
+pending_timeout(pending, now)
+	struct pending *pending;
+	struct timeval *now;
+{
+	int dirty = 0;
 
+	if (now->tv_sec - pending->p_tv.tv_sec > conf.c_timeout) {
+		if (conf.c_debug || conf.c_logexpired) {
+			mg_log(LOG_DEBUG,
+			    "(local): %s from %s to %s: greylisted "
+			    "entry timed out",
+			    pending->p_addr, pending->p_from,
+			    pending->p_rcpt);
+		}
+		pending_put(pending);
+		dirty = 1;
+	}
+
+	return dirty;
+}
+
+/* pending_lock must be locked */
 struct pending *
-pending_get(sa, salen, from, rcpt, date)  /* pending_lock must be write-locked */
+pending_get(sa, salen, from, rcpt, date)  
 	struct sockaddr *sa;
 	socklen_t salen;
 	char *from;
@@ -105,20 +141,13 @@ pending_get(sa, salen, from, rcpt, date)  /* pending_lock must be write-locked *
 {
 	struct pending *pending;
 	struct timeval tv;
-	int delay = conf.c_delay;
 	char addr[IPADDRSTRLEN];
 
 	if ((pending = malloc(sizeof(*pending))) == NULL)
 		goto out;
 
 	bzero((void *)pending, sizeof(pending));
-
-	if (date == 0) {
-		gettimeofday(&pending->p_tv, NULL);
-		pending->p_tv.tv_sec += delay;
-	} else {
-		pending->p_tv.tv_sec = date;
-	}
+	pending->p_tv.tv_sec = date;
 
 	if ((pending->p_sa = malloc(salen)) == NULL) {
 		free(pending);
@@ -152,38 +181,43 @@ pending_get(sa, salen, from, rcpt, date)  /* pending_lock must be write-locked *
 
 	pending->p_refcnt = 1;
 
-	TAILQ_INSERT_TAIL(&pending_head, pending, p_list); 
-
-	dump_dirty++;
+	TAILQ_INSERT_TAIL(&pending_head, pending, p_list);
+	TAILQ_INSERT_TAIL(&pending_buckets[BUCKET_HASH(pending->p_sa, 
+	    from, rcpt, PENDING_BUCKETS)].b_pending_head, pending, pb_list); 
 
 	(void)gettimeofday(&tv, NULL);
 
 	if (conf.c_debug) {
-		syslog(LOG_DEBUG, "created: %s from %s to %s delayed for %lds",
+		mg_log(LOG_DEBUG, "created: %s from %s to %s delayed for %lds",
 		    pending->p_addr, pending->p_from, pending->p_rcpt, 
 		    pending->p_tv.tv_sec - tv.tv_sec);
 	}
+
 out:
 	return pending;
 }
 
+/* pending_lock must be locked */
 void
-pending_put(pending) /* pending list should be write-locked */
+pending_put(pending) 
 	struct pending *pending;
 {
 	if (conf.c_debug) {
-		syslog(LOG_DEBUG, "removed: %s from %s to %s",
+		mg_log(LOG_DEBUG, "removed: %s from %s to %s",
 		    pending->p_addr, pending->p_from, pending->p_rcpt);
 	}
 
-	TAILQ_REMOVE(&pending_head, pending, p_list);	
+	TAILQ_REMOVE(&pending_head, pending, p_list);
+	TAILQ_REMOVE(&pending_buckets[BUCKET_HASH(pending->p_sa, 
+	    pending->p_from, pending->p_rcpt, PENDING_BUCKETS)].b_pending_head,
+	    pending, pb_list); 
+	
 	pending_free(pending);
 
-	dump_dirty++;
-
 	return;
 }
 
+
 void
 pending_del(sa, salen, from, rcpt, time)
 	struct sockaddr *sa;
@@ -195,15 +229,24 @@ pending_del(sa, salen, from, rcpt, time)
 	char addr[IPADDRSTRLEN];
 	struct pending *pending;
 	struct pending *next;
-	struct timeval tv;
+	struct timeval now;
+	struct pending_bucket *b;
+	int dirty = 0;
 
-	gettimeofday(&tv, NULL);
+	gettimeofday(&now, NULL);
 	if (!iptostring(sa, salen, addr, sizeof(addr)))
 		return;
 
-	PENDING_WRLOCK;	/* XXX take it as read and upgrade it */
-	for (pending = TAILQ_FIRST(&pending_head); pending; pending = next) {
-		next = TAILQ_NEXT(pending, p_list);
+	b = &pending_buckets[BUCKET_HASH(sa, from, rcpt, PENDING_BUCKETS)];
+	PENDING_LOCK;
+	for (pending = TAILQ_FIRST(&b->b_pending_head); 
+	    pending; pending = next) {
+		next = TAILQ_NEXT(pending, pb_list);
+
+		if (pending_timeout(pending, &now)) {
+			++dirty;
+			continue;
+		}
 
 		/*
 		 * Look for our entry.
@@ -213,30 +256,20 @@ pending_del(sa, salen, from, rcpt, time)
 		    (strcmp(rcpt, pending->p_rcpt) == 0) &&
 		    (pending->p_tv.tv_sec == time)) {
 			pending_put(pending);
+			++dirty;
 			break;
 		}
 
-		/*
-		 * Check for expired entries
-		 */
-		if (tv.tv_sec - pending->p_tv.tv_sec > conf.c_timeout) {
-			if (conf.c_debug) {
-				syslog(LOG_DEBUG,
-				    "del: %s from %s to %s timed out",
-				    pending->p_addr, pending->p_from,
-				    pending->p_rcpt);
-			}
-
-			pending_put(pending);
-			continue;
-		}
 	}
 	PENDING_UNLOCK;
+	
+	dump_touch(dirty);
+	
 	return;
 }
 
 int
-pending_check(sa, salen, from, rcpt, remaining, elapsed, queueid)
+pending_check(sa, salen, from, rcpt, remaining, elapsed, queueid, delay, aw)
 	struct sockaddr *sa;
 	socklen_t salen;
 	char *from;
@@ -244,46 +277,42 @@ pending_check(sa, salen, from, rcpt, remaining, elapsed, queueid)
 	time_t *remaining;
 	time_t *elapsed;
 	char *queueid;
+	time_t delay;
+	time_t aw;
 {
 	char addr[IPADDRSTRLEN];
 	struct pending *pending;
 	struct pending *next;
+	struct timeval tv;
 	time_t now;
-	time_t rest = -1;
-	time_t accepted = -1;
+	time_t rest;
+	time_t accepted;
 	int dirty = 0;
-	int delay = conf.c_delay;
+	struct pending_bucket *b;
 	ipaddr *mask = NULL;
+	time_t date;
 
-	now = time(NULL);
+	(void)gettimeofday(&tv, NULL);
+	now = tv.tv_sec;
 	if (!iptostring(sa, salen, addr, sizeof(addr)))
 		return 1;
 
-	PENDING_WRLOCK;	/* XXX take a read lock and upgrade */
-	for (pending = TAILQ_FIRST(&pending_head); pending; pending = next) {
-		next = TAILQ_NEXT(pending, p_list);
+	b = &pending_buckets[BUCKET_HASH(sa, from, rcpt, PENDING_BUCKETS)];
+	PENDING_LOCK;
+	for (pending = TAILQ_FIRST(&b->b_pending_head); 
+	    pending; pending = next) {
+		next = TAILQ_NEXT(pending, pb_list);
+
+		if (pending_timeout(pending, &tv)) {
+			++dirty;
+			continue;
+		}
 
 		/*
 		 * The time the entry shall be accepted
 		 */
 		accepted = pending->p_tv.tv_sec;
 
-		/*
-		 * Check for expired entries
-		 */
-		if (now - accepted > conf.c_timeout) {
-			if (conf.c_debug) {
-				syslog(LOG_DEBUG,
-				    "check: %s from %s to %s timed out",
-				    pending->p_addr, pending->p_from,
-				    pending->p_rcpt);
-			}
-
-			pending_put(pending);
-			dirty = 1;
-			continue;
-		}
-
 		/*
 		 * Look for our entry.
 		 */
@@ -303,12 +332,13 @@ pending_check(sa, salen, from, rcpt, remaining, elapsed, queueid)
 			rest = accepted - now;
 
 			if (rest <= 0) {
-				peer_delete(pending);
+				date = now + aw;
+				peer_delete(pending, date);
 				pending_put(pending);
-				autowhite_add(sa, salen, from, rcpt, NULL,
-				    queueid);
+				autowhite_add(sa, salen, from, rcpt, 
+				    &date, queueid);
 				rest = 0;
-				dirty = 1;
+				++dirty;
 			}
 
 			goto out;
@@ -319,24 +349,30 @@ pending_check(sa, salen, from, rcpt, remaining, elapsed, queueid)
 	 * It was not found. Create it and propagagte it to peers.
 	 * Error handling is useless here, we will tempfail anyway
 	 */
-	pending = pending_get(sa, salen, from, rcpt, (time_t)0);
-	peer_create(pending);
-	rest = delay;
-	dirty = 1;
+	accepted = now + delay;
+	rest = 0;
+	pending = pending_get(sa, salen, from, rcpt, accepted);
+	if (pending) {
+		++dirty;
+		peer_create(pending);
+		rest = pending->p_tv.tv_sec - now;
+	} /* otherwise return 1 and accept the mail */
 
 out:
 	PENDING_UNLOCK;
 
 	if (remaining != NULL)
-		*remaining = rest; 
+		*remaining = rest;
 
 	if (elapsed != NULL)
 		*elapsed = now - (accepted - delay);
 
-	if (dirty)
+	if (dirty) {
+		dump_touch(dirty);
 		dump_flush();
+	}
 
-	if (rest == 0)
+	if (rest <= 0)
 		return 1;
 	else
 		return 0;
@@ -346,23 +382,42 @@ int
 pending_textdump(stream)
 	FILE *stream;
 {
+	struct timeval now;
 	struct pending *pending;
+	struct pending *next;
 	int done = 0;
 	char textdate[DATELEN + 1];
 	struct tm tm;
 
+	gettimeofday(&now, NULL);
+
 	fprintf(stream, "\n\n#\n# greylisted tuples\n#\n");
 	fprintf(stream, "# Sender IP\t%s\t%s\tTime accepted\n", 
 	    "Sender e-mail", "Recipient e-mail");
 
-	PENDING_RDLOCK;
-	TAILQ_FOREACH(pending, &pending_head, p_list) {
-		localtime_r((time_t *)&pending->p_tv.tv_sec, &tm);
-		strftime(textdate, DATELEN, "%Y-%m-%d %T", &tm);
+	PENDING_LOCK;
+	for (pending = TAILQ_FIRST(&pending_head); pending; pending = next) {
+		next = TAILQ_NEXT(pending, p_list);
 
-		fprintf(stream, "%s\t%s\t%s\t%ld # %s\n", 
-		    pending->p_addr, pending->p_from, 
-		    pending->p_rcpt, (long)pending->p_tv.tv_sec, textdate);
+		if (pending_timeout(pending, &now))
+			continue;
+
+		if (conf.c_dump_no_time_translation) {
+			fprintf(stream, "%s\t%s\t%s\t%ld\n", 
+			    pending->p_addr, pending->p_from, 
+			    pending->p_rcpt, (long)pending->p_tv.tv_sec);
+		} else {
+			time_t ti;
+
+			ti = pending->p_tv.tv_sec;
+			localtime_r(&ti, &tm);
+			strftime(textdate, DATELEN, "%Y-%m-%d %T", &tm);
+		
+			fprintf(stream, "%s\t%s\t%s\t%ld # %s\n", 
+			    pending->p_addr, pending->p_from, 
+			    pending->p_rcpt, 
+			    (long)pending->p_tv.tv_sec, textdate);
+		}
 		
 		done++;
 	}
@@ -375,9 +430,9 @@ struct pending *
 pending_ref(pending)
 	struct pending *pending;
 {
-	WRLOCK(refcnt_lock);
+	pthread_mutex_lock(&refcnt_lock);
 	pending->p_refcnt++;
-	UNLOCK(refcnt_lock);
+	pthread_mutex_unlock(&refcnt_lock);
 	return pending;
 }
 
@@ -385,13 +440,13 @@ void
 pending_free(pending)
 	struct pending *pending;
 {
-	WRLOCK(refcnt_lock);
+	pthread_mutex_lock(&refcnt_lock);
 	pending->p_refcnt--;
 	if (pending->p_refcnt > 0) {
-		UNLOCK(refcnt_lock);
+		pthread_mutex_unlock(&refcnt_lock);
 		return;
 	}
-	UNLOCK(refcnt_lock);
+	pthread_mutex_unlock(&refcnt_lock);
 	free(pending->p_sa);
 	free(pending->p_addr);
 	free(pending->p_from);
@@ -419,7 +474,7 @@ ip_match(sa, pat, mask)
 		break;
 #ifdef AF_INET6
 	case AF_INET6:
-#ifdef HAVE_GETADDRINFO
+#if defined(HAVE_SIN6_SCOPE_ID) && defined(HAVE_GETADDRINFO)
 		if (SA6(pat)->sin6_scope_id != 0 &&
 		    SA6(sa)->sin6_scope_id != SA6(pat)->sin6_scope_id)
 			return 0;
@@ -453,7 +508,7 @@ ip_equal(sa, pat)
 		break;
 #ifdef AF_INET6
 	case AF_INET6:
-#ifdef HAVE_GETADDRINFO
+#if defined(HAVE_SIN6_SCOPE_ID) && defined(HAVE_GETADDRINFO)
 		if (SA6(pat)->sin6_scope_id != 0 &&
 		    SA6(sa)->sin6_scope_id != SA6(pat)->sin6_scope_id)
 			return 0;
@@ -576,3 +631,51 @@ ipfromstring(str, sa, salen, family)
 	return 0;
 #endif
 }
+
+void
+pending_del_addr(sa, salen, queueid, acl_line)
+	struct sockaddr *sa;
+	socklen_t salen;
+	char *queueid;
+	int acl_line;
+{
+	char addr[IPADDRSTRLEN];
+	struct pending *pending;
+	struct pending *next;
+	int sync_flush_done = 0;
+	int count_pending = 0;
+	int count_white = 0;
+        char aclstr[16];
+
+	if (!iptostring(sa, salen, addr, sizeof(addr)))
+		return;
+
+	PENDING_LOCK;
+	for (pending = TAILQ_FIRST(&pending_head); pending; pending = next) {
+		next = TAILQ_NEXT(pending, p_list);
+
+		if (strncmp(addr, pending->p_addr, sizeof(addr)) == 0) {
+			if (!sync_flush_done) {
+				peer_flush(pending);
+				sync_flush_done = 1;
+			}
+			pending_put(pending);
+			count_pending++;
+		}
+
+	}
+	PENDING_UNLOCK;
+	dump_touch(count_pending);
+
+	/* And flush autowhite list as well... */
+	count_white = autowhite_del_addr(sa, salen);
+	
+	if (queueid != NULL) {
+		*aclstr = '\0';
+        	if (acl_line != 0)
+			snprintf(aclstr, sizeof(aclstr), " (ACL %d)", acl_line);
+		mg_log(LOG_INFO, "%s: addr %s flushed, removed %d grey and %d autowhite%s",
+			queueid, addr, count_pending, count_white, aclstr);
+	}
+	return;
+}
