@@ -1,4 +1,4 @@
-/* $Id: spf.h,v 1.8 2004/08/01 09:27:03 manu Exp $ */
+/* $Id: spf.h,v 1.13 2007/03/22 23:11:35 manu Exp $ */
 
 /*
  * Copyright (c) 2004 Emmanuel Dreyfus
@@ -39,18 +39,15 @@
 #include <arpa/inet.h>
 #include <netinet/in.h>
 
-#include "except.h"
-
-#if defined(HAVE_SPF_ALT)
-int spf_alt_check(struct sockaddr *, socklen_t, char *, char *);
-#define SPF_CHECK(w,x,y,z) spf_alt_check((w),(x),(y),(z))
-
-#elif defined(HAVE_SPF)
-int spf_check(struct sockaddr *, socklen_t, char*, char *);
-#define SPF_CHECK(w,x,y,z) spf_check((w),(x),(y),(z))
+#include "acl.h"
 
+#if (defined(HAVE_SPF) || defined(HAVE_SPF_ALT) || \
+     defined(HAVE_SPF2_10) || defined(HAVE_SPF2))
+int spf_check(acl_data_t *, acl_stage_t,
+		  struct acl_param *, struct mlfi_priv *);
+#define SPF_CHECK(priv) spf_check(NULL, AS_RCPT, NULL, (priv))
 #else
-#define SPF_CHECK(w,x,y,z) EXF_NONE
+#define SPF_CHECK(priv) 0
 
 #endif
 
