@@ -1,7 +1,7 @@
-/* $Id: sync.c,v 1.56 2005/10/20 07:24:53 manu Exp $ */
+/* $Id: sync.c,v 1.80 2007/10/05 23:12:47 manu Exp $ */
 
 /*
- * Copyright (c) 2004 Emmanuel Dreyfus
+ * Copyright (c) 2004-2007 Emmanuel Dreyfus
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,7 +34,7 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID
-__RCSID("$Id: sync.c,v 1.56 2005/10/20 07:24:53 manu Exp $");
+__RCSID("$Id: sync.c,v 1.80 2007/10/05 23:12:47 manu Exp $");
 #endif
 #endif
 
@@ -59,22 +59,64 @@ __RCSID("$Id: sync.c,v 1.56 2005/10/20 07:24:53 manu Exp $");
 #include "autowhite.h"
 #include "milter-greylist.h"
 
+#ifdef USE_DMALLOC
+#include <dmalloc.h> 
+#endif
+
+#define SYNC_PROTO_CURRENT 3
+
 struct sync_master_sock {
 	int runs;
 	int sock;
 };
 
+/* errors returned by stdio that should not cause an exit */
+static int sync_ignored_errno[] = { 
+#ifdef EAGAIN
+	EAGAIN,
+#endif /* EAGAIN */
+#ifdef ECONNABORTED
+	ECONNABORTED,
+#endif /* ECONNABORTED */
+#ifdef EMFILE
+	EMFILE,
+#endif /* EMFILE */
+#ifdef ENFILE
+	ENFILE,
+#endif /* ENFILE */
+#ifdef ENOBUFS
+	ENOBUFS,
+#endif /* ENOBUFS */
+#ifdef ENOMEM
+	ENOMEM,
+#endif /* ENOMEM */
+#ifdef ENOSR
+	ENOSR,
+#endif /* ENOSR */
+#ifdef EWOULDBLOCK
+	EWOULDBLOCK,
+#endif /* EWOULDBLOCK */
+	0,
+};
+
+static pthread_mutex_t sync_master_lock = PTHREAD_MUTEX_INITIALIZER;
 struct sync_master_sock sync_master4 = { 0, -1 };
 struct sync_master_sock sync_master6 = { 0, -1 };
+
 struct peerlist peer_head;
 pthread_rwlock_t peer_lock; /* For the peer list */
-pthread_rwlock_t sync_lock; /* For all peer's sync queue */
-pthread_cond_t sync_sleepflag;
+
+static pthread_mutex_t sync_dirty_lock = PTHREAD_MUTEX_INITIALIZER;
+static pthread_cond_t sync_sleepflag = PTHREAD_COND_INITIALIZER;
+static int sync_dirty = 0;
 
 static void sync_listen(char *, char *, struct sync_master_sock *);
 static int local_addr(const struct sockaddr *sa, const socklen_t salen);
 static int sync_queue_poke(struct peer *, struct sync *);
 static struct sync * sync_queue_peek(struct peer *);
+static int select_protocol(struct peer *, int, FILE *);
+static void sync_vers(FILE *, int);
+static int is_fatal(int);
 
 void
 peer_init(void) {
@@ -82,23 +124,11 @@ peer_init(void) {
 
 	LIST_INIT(&peer_head);
 	if ((error = pthread_rwlock_init(&peer_lock, NULL)) != 0) {
-		syslog(LOG_ERR, 
+		mg_log(LOG_ERR, 
 		    "pthread_rwlock_init failed: %s", strerror(error));
 		exit(EX_OSERR);
 	}
 
-	if ((error = pthread_rwlock_init(&sync_lock, NULL)) != 0) {
-		syslog(LOG_ERR, 
-		    "pthread_rwlock_init failed: %s", strerror(error));
-		exit(EX_OSERR);
-	}
-
-	if ((error = pthread_cond_init(&sync_sleepflag, NULL)) != 0) {
-		syslog(LOG_ERR, 
-		    "pthread_cond_init failed: %s", strerror(error));
-		exit(EX_OSERR);
-	}
-
 	return;
 }
 
@@ -116,9 +146,10 @@ peer_clear(void) {
 			sync_free(sync);
 			
 		if (peer->p_stream != NULL)
-			fclose(peer->p_stream);
+			Fclose(peer->p_stream);
 
 		LIST_REMOVE(peer, p_list);
+		pthread_mutex_destroy(&peer->p_mtx);
 		free(peer->p_name);
 		free(peer);
 	}
@@ -136,7 +167,7 @@ peer_add(peername)
 
 	if ((peer = malloc(sizeof(*peer))) == NULL ||
 	    (peer->p_name = strdup(peername)) == NULL) {
-		syslog(LOG_ERR, "cannot add peer: %s", strerror(errno));
+		mg_log(LOG_ERR, "cannot add peer: %s", strerror(errno));
 		exit(EX_OSERR);
 	}
 
@@ -144,13 +175,14 @@ peer_add(peername)
 	peer->p_stream = NULL;
 	peer->p_flags = 0;
 	TAILQ_INIT(&peer->p_deferred);
+	pthread_mutex_init(&peer->p_mtx, NULL);
 
 	PEER_WRLOCK;
 	LIST_INSERT_HEAD(&peer_head, peer, p_list);
 	PEER_UNLOCK;
 
 	if (conf.c_debug)
-		printf("load peer %s\n", peer->p_name);
+		mg_log(LOG_DEBUG, "load peer %s", peer->p_name);
 
 	return;
 }
@@ -166,7 +198,7 @@ peer_create(pending)
 		goto out;
 
 	LIST_FOREACH(peer, &peer_head, p_list) 
-		sync_queue(peer, PS_CREATE, pending);
+		sync_queue(peer, PS_CREATE, pending, -1); /* -1: unused */
 
 out:
 	PEER_UNLOCK;
@@ -175,8 +207,9 @@ out:
 }
 
 void
-peer_delete(pending)
+peer_delete(pending, autowhite)
 	struct pending *pending;
+	time_t autowhite;
 {
 	struct peer *peer;
 
@@ -185,7 +218,7 @@ peer_delete(pending)
 		goto out;
 
 	LIST_FOREACH(peer, &peer_head, p_list) 
-		sync_queue(peer, PS_DELETE, pending);
+		sync_queue(peer, PS_DELETE, pending, autowhite);
 
 out:
 	PEER_UNLOCK;
@@ -198,16 +231,16 @@ sync_queue_poke(peer, sync)
 	struct peer *peer;
 	struct sync *sync;
 {
-	SYNC_WRLOCK;
+	int r = 0;
+
+	pthread_mutex_lock(&peer->p_mtx);
 	if (peer->p_qlen < SYNC_MAXQLEN) {
 		TAILQ_INSERT_HEAD(&peer->p_deferred, sync, s_list);
 		peer->p_qlen++;
-		SYNC_UNLOCK;
-		return 1;
-	} else {
-		SYNC_UNLOCK;
-		return 0;
+		r = 1;
 	}
+	pthread_mutex_unlock(&peer->p_mtx);
+	return r;
 }
 
 static struct sync *
@@ -216,50 +249,106 @@ sync_queue_peek(peer)
 {
 	struct sync *sync;
 
-	SYNC_WRLOCK;
+	pthread_mutex_lock(&peer->p_mtx);
 	sync = TAILQ_FIRST(&peer->p_deferred);
 	if (!TAILQ_EMPTY(&peer->p_deferred)) {
 		TAILQ_REMOVE(&peer->p_deferred, sync, s_list);
 		peer->p_qlen--;
 	}
-	SYNC_UNLOCK;
+	pthread_mutex_unlock(&peer->p_mtx);
 	return sync;
 }
 
 int
-sync_send(peer, type, pending)	/* peer list is read-locked */
+sync_send(peer, type, pending, autowhite) /* peer list is read-locked */
 	struct peer *peer;
 	peer_sync_t type;
 	struct pending *pending;
+	time_t autowhite;
 {
 	char sep[] = " \n\t\r";
 	char *replystr;
 	int replycode;
 	char line[LINELEN + 1];
 	char *cookie = NULL;
+	char *keyw;
+	char awstr[LINELEN + 1];
+	int bw;
 
 	if ((peer->p_stream == NULL) && (peer_connect(peer) != 0))
 		return -1;
 
-	if (type == PS_CREATE)
-		fprintf(peer->p_stream, "add ");
-	else
-		fprintf(peer->p_stream, "del ");
+	*line = '\0';
+	switch(type) {
+	case PS_FLUSH:
+		bw = snprintf(line, LINELEN, "flush addr %s\r\n",
+		    pending->p_addr);
+		break;
+	case PS_CREATE:
+		bw = snprintf(line, LINELEN, "add addr %s from %s "
+		    "rcpt %s date %ld\r\n", pending->p_addr, 
+			pending->p_from, pending->p_rcpt, 
+			(long)pending->p_tv.tv_sec);
+		break;
+	default:
+		if (peer->p_vers >= 2) {
+			keyw = "del2";
+			snprintf(awstr, LINELEN, " aw %ld", (long)autowhite);
+		} else {
+			keyw = "del";
+			awstr[0] = '\0';
+		}
+		bw = snprintf(line, LINELEN, "%s addr %s from %s "
+		    "rcpt %s date %ld%s\r\n", keyw, pending->p_addr, 
+			pending->p_from, pending->p_rcpt, 
+			(long)pending->p_tv.tv_sec, awstr);
+		break;
+	}
+
+	if (bw > LINELEN) {
+		mg_log(LOG_ERR, "closing connexion with peer %s: "
+		    "send buffer would overflow (%d entries queued)", 
+		    peer->p_name, peer->p_qlen);
+		Fclose(peer->p_stream);
+		peer->p_stream = NULL;
+		return -1;
+	}
+
+	bw = fprintf(peer->p_stream, "%s", line);
+	if (bw != strlen(line)) {
+		mg_log(LOG_ERR, "closing connexion with peer %s: "
+		    "%s (%d entries queued) - I was unable to send "
+		    "complete line \"%s\" - bytes written: %i", 
+		    peer->p_name, strerror(errno), peer->p_qlen, 
+		    line, bw);
+		Fclose(peer->p_stream);
+		peer->p_stream = NULL;
+		return -1;
+	}
 
-	fprintf(peer->p_stream, "addr %s from %s rcpt %s date %ld\r\n", 
-	    pending->p_addr, pending->p_from, 
-	    pending->p_rcpt, (long)pending->p_tv.tv_sec);
 	fflush(peer->p_stream);
 
 	/* 
 	 * Check the return code 
 	 */
+	get_more:
 	sync_waitdata(peer->p_socket);
 	if (fgets(line, LINELEN, peer->p_stream) == NULL) {
-		syslog(LOG_ERR, "lost connexion with peer %s: "
+		if (errno == EAGAIN) {
+			if ( feof(peer->p_stream) ) {
+				mg_log(LOG_ERR, "lost connexion with peer %s: "
+		  		  "%s (%d entries queued)", 
+				    peer->p_name, strerror(errno), peer->p_qlen);
+				Fclose(peer->p_stream);
+				peer->p_stream = NULL;
+				return -1;
+			}
+			goto get_more;
+		}
+		mg_log(LOG_ERR, "lost connexion with peer %s: "
 		    "%s (%d entries queued)", 
 		    peer->p_name, strerror(errno), peer->p_qlen);
-		fclose(peer->p_stream);
+		Fclose(peer->p_stream);
 		peer->p_stream = NULL;
 		return -1;
 	}
@@ -272,26 +361,26 @@ sync_send(peer, type, pending)	/* peer list is read-locked */
 	fflush(peer->p_stream);
 
 	if ((replystr = strtok_r(line, sep, &cookie)) == NULL) {
-		syslog(LOG_ERR, "Unexpected reply \"%s\" from %s, "
+		mg_log(LOG_ERR, "Unexpected reply \"%s\" from %s, "
 		    "closing connexion (%d entries queued)", 
 		    line, peer->p_name, peer->p_qlen);
-		fclose(peer->p_stream);
+		Fclose(peer->p_stream);
 		peer->p_stream = NULL;
 		return -1;
 	}
 
 	replycode = atoi(replystr);
 	if (replycode != 201) {
-		syslog(LOG_ERR, "Unexpected reply \"%s\" from %s, "
+		mg_log(LOG_ERR, "Unexpected reply \"%s\" from %s, "
 		    "closing connexion (%d entries queued)", 
 		    line, peer->p_name, peer->p_qlen);
-		fclose(peer->p_stream);
+		Fclose(peer->p_stream);
 		peer->p_stream = NULL;
 		return -1;
 	}
 
 	if (conf.c_debug)
-		syslog(LOG_DEBUG, "sync one entry with %s", peer->p_name);
+		mg_log(LOG_DEBUG, "sync one entry with %s", peer->p_name);
 
 	return 0;
 }
@@ -324,7 +413,7 @@ peer_connect(peer)	/* peer list is read-locked */
 	char *cookie = NULL;
 
 	if (peer->p_stream != NULL)
-		syslog(LOG_ERR, "peer_connect called and peer->p_stream != 0");
+		mg_log(LOG_ERR, "peer_connect called and peer->p_stream != 0");
 
 	if (conf.c_syncport != NULL) {
 		service = htons(atoi(conf.c_syncport));
@@ -340,7 +429,7 @@ peer_connect(peer)	/* peer list is read-locked */
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = SOCK_STREAM;
 	if ((err = getaddrinfo(peer->p_name, "0", &hints, &res0)) != 0) {
-		syslog(LOG_ERR, "cannot sync with peer %s, "
+		mg_log(LOG_ERR, "cannot sync with peer %s, "
 		    "getaddrinfo failed: %s (%d entries queued)",
 		    peer->p_name, gai_strerror(err), peer->p_qlen);
 		return -1;
@@ -368,7 +457,11 @@ peer_connect(peer)	/* peer list is read-locked */
 		switch (res->ai_family) {
 		case AF_INET:
 			SA4(res->ai_addr)->sin_port = service;
-			laddrstr = "0.0.0.0";
+			if (conf.c_syncsrcaddr != NULL) {
+				laddrstr = conf.c_syncsrcaddr;
+			} else {
+				laddrstr = "0.0.0.0";
+				}
 			break;
 #ifdef AF_INET6
 		case AF_INET6:
@@ -377,7 +470,7 @@ peer_connect(peer)	/* peer list is read-locked */
 			break;
 #endif
 		default:
-			syslog(LOG_ERR, "cannot sync, unknown address family");
+			mg_log(LOG_ERR, "cannot sync, unknown address family");
 			close(s);
 			s = -1;
 			continue;
@@ -393,7 +486,7 @@ peer_connect(peer)	/* peer list is read-locked */
 	}
 	freeaddrinfo(res0);
 	if (s < 0) {
-		syslog(LOG_ERR,
+		mg_log(LOG_ERR,
 		    "cannot sync with peer %s: %s (%d entries queued)",
 		    peer->p_name, strerror(errno), peer->p_qlen);
 		return -1;
@@ -402,7 +495,7 @@ peer_connect(peer)	/* peer list is read-locked */
 	raddrlen = sizeof(raddr);
 	if (ipfromstring(peer->p_name, SA(&raddr), &raddrlen,
 	    AF_UNSPEC) != 1) {
-		syslog(LOG_ERR, "cannot sync, invalid address");
+		mg_log(LOG_ERR, "cannot sync, invalid address");
 		return -1;
 	}
 
@@ -414,7 +507,11 @@ peer_connect(peer)	/* peer list is read-locked */
 	switch (SA(&raddr)->sa_family) {
 	case AF_INET:
 		SA4(&raddr)->sin_port = service;
-		laddrstr = "0.0.0.0";
+		if (conf.c_syncsrcaddr != NULL) {
+			laddrstr = conf.c_syncsrcaddr;
+		} else {
+			laddrstr = "0.0.0.0";
+			}
 		break;
 #ifdef AF_INET6
 	case AF_INET6:
@@ -423,7 +520,7 @@ peer_connect(peer)	/* peer list is read-locked */
 		break;
 #endif
 	default:
-		syslog(LOG_ERR, "cannot sync, unknown address family");
+		mg_log(LOG_ERR, "cannot sync, unknown address family");
 		return -1;
 	}
 
@@ -433,7 +530,7 @@ peer_connect(peer)	/* peer list is read-locked */
 		proto = pe->p_proto;
 
 	if ((s = socket(SA(&raddr)->sa_family, SOCK_STREAM, proto)) == -1) {
-		syslog(LOG_ERR, "cannot sync with peer %s, "
+		mg_log(LOG_ERR, "cannot sync with peer %s, "
 		    "socket failed: %s (%d entries queued)", 
 		    peer->p_name, strerror(errno), peer->p_qlen);
 		return -1;
@@ -442,13 +539,13 @@ peer_connect(peer)	/* peer list is read-locked */
 	laddrlen = sizeof(laddr);
 	if (ipfromstring(laddrstr, SA(&laddr), &laddrlen,
 	    SA(&raddr)->sa_family) != 1) {
-		syslog(LOG_ERR, "cannot sync, invalid address");
+		mg_log(LOG_ERR, "cannot sync, invalid address");
 		close(s);
 		return -1;
 	}
 
 	if (bind(s, SA(&laddr), laddrlen) != 0) {
-		syslog(LOG_ERR, "cannot sync with peer %s, "
+		mg_log(LOG_ERR, "cannot sync with peer %s, "
 		    "bind failed: %s (%d entries queued)",
 		    peer->p_name, strerror(errno), peer->p_qlen);
 		close(s);
@@ -456,7 +553,7 @@ peer_connect(peer)	/* peer list is read-locked */
 	}
 
 	if (connect(s, SA(&raddr), raddrlen) != 0) {
-		syslog(LOG_ERR, "cannot sync with peer %s, "
+		mg_log(LOG_ERR, "cannot sync with peer %s, "
 		    "connect failed: %s (%d entries queued)", 
 		    peer->p_name, strerror(errno), peer->p_qlen);
 		close(s);
@@ -465,26 +562,33 @@ peer_connect(peer)	/* peer list is read-locked */
 #endif
 
 	param = O_NONBLOCK;
-	if (fcntl(s, F_SETFL, &param) != 0) {
-		syslog(LOG_ERR, "cannot set non blocking I/O with %s: %s",
+	if (fcntl(s, F_SETFL, param) != 0) {
+		mg_log(LOG_ERR, "cannot set non blocking I/O with %s: %s",
 		    peer->p_name, strerror(errno));
 	}
 
-	if ((stream = fdopen(s, "w+")) == NULL) {
-		syslog(LOG_ERR, "cannot sync with peer %s, "
+	errno = 0;
+	if ((stream = Fdopen(s, "w+")) == NULL) {
+		mg_log(LOG_ERR, "cannot sync with peer %s, "
 		    "fdopen failed: %s (%d entries queued)", 
-		    peer->p_name, strerror(errno), peer->p_qlen);
+		    peer->p_name, 
+		    (errno == 0) ? "out of stdio streams" : strerror(errno),
+		    peer->p_qlen);
 		close(s);
 		return -1;
 	}
 
+#ifdef USE_FD_POOL
+	s = fileno(stream);     /* the socket descriptor could have been replaced by Fdopen() ! */
+#endif
+
 	if (setvbuf(stream, NULL, _IOLBF, 0) != 0)
-		syslog(LOG_ERR, "cannot set line buffering with peer %s: %s", 
+		mg_log(LOG_ERR, "cannot set line buffering with peer %s: %s", 
 		    peer->p_name, strerror(errno));
 
 	sync_waitdata(s);	
 	if (fgets(line, LINELEN, stream) == NULL) {
-		syslog(LOG_ERR, "Lost connexion with peer %s: "
+		mg_log(LOG_ERR, "Lost connexion with peer %s: "
 		    "%s (%d entries queued)", 
 		    peer->p_name, strerror(errno), peer->p_qlen);
 		goto bad;
@@ -498,7 +602,7 @@ peer_connect(peer)	/* peer list is read-locked */
 	fflush(stream);
 
 	if ((replystr = strtok_r(line, sep, &cookie)) == NULL) {
-		syslog(LOG_ERR, "Unexpected reply \"%s\" from peer %s "
+		mg_log(LOG_ERR, "Unexpected reply \"%s\" from peer %s "
 		    "closing connexion (%d entries queued)", 
 		    line, peer->p_name, peer->p_qlen);
 		goto bad;
@@ -506,24 +610,42 @@ peer_connect(peer)	/* peer list is read-locked */
 
 	replycode = atoi(replystr);
 	if (replycode != 200) {
-		syslog(LOG_ERR, "Unexpected reply \"%s\" from peer %s "
+		mg_log(LOG_ERR, "Unexpected reply \"%s\" from peer %s "
 		    "closing connexion (%d entries queued)", 
 		    line, peer->p_name, peer->p_qlen);
 		goto bad;
 	}
 
-	syslog(LOG_INFO, "Connection to %s established", peer->p_name);
+	if ((peer->p_vers = select_protocol(peer, s, stream)) == 0)
+		goto bad;	
+
+	mg_log(LOG_INFO, "Connection to %s established, protocol version %d", 
+	    peer->p_name, peer->p_vers);
 	peer->p_stream = stream;
 	peer->p_socket = s;
+
 	return 0;
 
 bad:
-	fclose(stream);
+	Fclose(stream);
 	peer->p_stream = NULL;
 
 	return -1;
 }
 
+void
+sync_master_stop(void) {
+	if (sync_master4.sock != 1) {
+		close(sync_master4.sock);
+		sync_master4.sock = -1;
+	}
+
+	if (sync_master6.sock != 1) {
+		close(sync_master6.sock);
+		sync_master6.sock = -1;
+	}
+}
+
 void
 sync_master_restart(void) {
 	pthread_t tid;
@@ -534,8 +656,9 @@ sync_master_restart(void) {
 	empty = LIST_EMPTY(&peer_head);
 	PEER_UNLOCK;
 
+	pthread_mutex_lock(&sync_master_lock);
 	if (empty || sync_master4.runs || sync_master6.runs)
-		return;
+		goto last;
 
 	if (conf.c_syncaddr != NULL) {
 		if (strchr(conf.c_syncaddr, ':'))
@@ -554,21 +677,21 @@ sync_master_restart(void) {
 
 
 	if (!sync_master4.runs && !sync_master6.runs) {
-		syslog(LOG_ERR, "cannot start MX sync, socket failed: %s",
+		mg_log(LOG_ERR, "cannot start MX sync, socket failed: %s",
 		    strerror(errno));
 		exit(EX_OSERR);
 	}
 	if (sync_master6.runs) {
 		if ((error = pthread_create(&tid, NULL, sync_master,
 		    (void *)&sync_master6)) != 0) {
-			syslog(LOG_ERR, 
-			    "Cannot run MX sync thread for IPv6: %s\n",
+			mg_log(LOG_ERR, 
+			    "Cannot run MX sync thread for IPv6: %s",
 			    strerror(error));
 			exit(EX_OSERR);
 		}
 		if ((error = pthread_detach(tid)) != 0) {
-			syslog(LOG_ERR, 
-			    "pthread_detach failed for IPv6 MX sync: %s\n",
+			mg_log(LOG_ERR, 
+			    "pthread_detach failed for IPv6 MX sync: %s",
 			    strerror(error));
 			exit(EX_OSERR);
 		}
@@ -576,18 +699,20 @@ sync_master_restart(void) {
 	if (sync_master4.runs) {
 		if ((error = pthread_create(&tid, NULL, sync_master,
 		    (void *)&sync_master4)) != 0) {
-			syslog(LOG_ERR, 
-			    "Cannot run MX sync thread for IPv4: %s\n",
+			mg_log(LOG_ERR, 
+			    "Cannot run MX sync thread for IPv4: %s",
 			    strerror(error));
 			exit(EX_OSERR);
 		}
 		if ((error = pthread_detach(tid)) != 0) {
-			syslog(LOG_ERR, 
-			    "pthread_detach failed for IPv4 MX sync: %s\n",
+			mg_log(LOG_ERR, 
+			    "pthread_detach failed for IPv4 MX sync: %s",
 			    strerror(error));
 			exit(EX_OSERR);
 		}
 	}
+last:
+	pthread_mutex_unlock(&sync_master_lock);
 }
 
 void *
@@ -596,6 +721,7 @@ sync_master(arg)
 {
 	struct sync_master_sock *sms = arg;
 
+	conf_retain();
 	for (;;) {
 		sockaddr_t raddr;
 		socklen_t raddrlen;
@@ -605,33 +731,47 @@ sync_master(arg)
 		struct peer *peer;
 		char peerstr[IPADDRSTRLEN];
 		int error;
+		int sock;
 
+		pthread_mutex_lock(&sync_master_lock);
+		sock = sms->sock;
+		pthread_mutex_unlock(&sync_master_lock);
+
+		/* TODO: accept connections in nonblocking mode
+		 * in order to watch conf change */
 		bzero((void *)&raddr, sizeof(raddr));
 		raddrlen = sizeof(raddr);
-		if ((fd = accept(sms->sock, SA(&raddr), &raddrlen)) == -1) {
-			syslog(LOG_ERR, "incoming connexion "
-			    "failed: %s\n", strerror(errno));
+		if ((fd = accept(sock, SA(&raddr), &raddrlen)) == -1) {
+			mg_log(LOG_ERR, "incoming connexion "
+			    "failed: %s", strerror(errno));
+
+			if (is_fatal(errno))
+                        	exit(EX_OSERR);
+
 
-			if (errno != ECONNABORTED)
-				exit(EX_OSERR);
-			continue;
 		}
 		unmappedaddr(SA(&raddr), &raddrlen);
 
+		conf_release();
+		conf_retain();
+
 		iptostring(SA(&raddr), raddrlen, peerstr, sizeof(peerstr));
-		syslog(LOG_INFO, "Incoming MX sync connexion from %s", 
+		mg_log(LOG_INFO, "Incoming MX sync connexion from %s", 
 		    peerstr);
 
-		if ((stream = fdopen(fd, "w+")) == NULL) {
-			syslog(LOG_ERR, 
+		errno = 0;
+		if ((stream = Fdopen(fd, "w+")) == NULL) {
+			mg_log(LOG_ERR, 
 			    "incoming connexion from %s failed, "
-			    "fdopen fail: %s", peerstr, strerror(errno));
+			    "fdopen fail: %s", peerstr, 
+		    	    (errno == 0) ? "out of stdio streams"
+					 : strerror(errno));
 			close(fd);
 			exit(EX_OSERR);
 		}
 
 		if (setvbuf(stream, NULL, _IOLBF, 0) != 0)
-			syslog(LOG_ERR, "cannot set line buffering: %s\n", 
+			mg_log(LOG_ERR, "cannot set line buffering: %s", 
 			    strerror(errno));	
 		
 		/*
@@ -643,10 +783,13 @@ sync_master(arg)
 			fprintf(stream, "105 No more peers, shutting down!\n");
 
 			PEER_UNLOCK;
-			fclose(stream);
+			Fclose(stream);
+			pthread_mutex_lock(&sync_master_lock);
 			close(sms->sock);
 			sms->sock = -1;
 			sms->runs = 0;
+			pthread_mutex_unlock(&sync_master_lock);
+			conf_release();
 			return NULL;
 		}
 			
@@ -662,7 +805,7 @@ sync_master(arg)
 			hints.ai_socktype = SOCK_STREAM;
 			err = getaddrinfo(peer->p_name, "0", &hints, &res0);
 			if (err != 0) {
-				syslog(LOG_ERR, "cannot resolve %s: %s",
+				mg_log(LOG_ERR, "cannot resolve %s: %s",
 				    peer->p_name, gai_strerror(err));
 				continue;
 			}
@@ -682,7 +825,7 @@ sync_master(arg)
 			addrlen = sizeof(addr);
 			if (ipfromstring(peer->p_name, SA(&addr), &addrlen,
 			     AF_UNSPEC) != 1) {
-				syslog(LOG_ERR, "cannot resolve %s",
+				mg_log(LOG_ERR, "cannot resolve %s",
 				    peer->p_name);
 				continue;
 			}
@@ -694,24 +837,24 @@ sync_master(arg)
 		PEER_UNLOCK;
 
 		if (peer == NULL) {
-			syslog(LOG_INFO, "Remote host %s is not a peer MX", 
+			mg_log(LOG_INFO, "Remote host %s is not a peer MX", 
 			    peerstr);
 			fprintf(stream, 
 			    "106 You have no permission to talk, go away!\n");
-			fclose(stream);
+			Fclose(stream);
 			continue;
 		}
 
 		if ((error = pthread_create(&tid, NULL, 
 		    (void *(*)(void *))sync_server, (void *)stream)) != 0) {
-			syslog(LOG_ERR, "incoming connexion from %s failed, "
+			mg_log(LOG_ERR, "incoming connexion from %s failed, "
 			    "pthread_create failed: %s", 
 			    peerstr, strerror(error));
-			fclose(stream);
+			Fclose(stream);
 			continue;
 		}
 		if ((error = pthread_detach(tid)) != 0) {
-			syslog(LOG_ERR, "incoming connexion from %s failed, "
+			mg_log(LOG_ERR, "incoming connexion from %s failed, "
 			    "pthread_detach failed: %s",
 			    peerstr, strerror(error));
 			exit(EX_OSERR);
@@ -719,10 +862,11 @@ sync_master(arg)
 	}
 
 	/* NOTREACHED */
-	syslog(LOG_ERR, "sync_master quitted unexpectedly");
+	mg_log(LOG_ERR, "sync_master quitted unexpectedly");
 	return NULL;
 }
 
+/* sync_master_lock must be locked */
 static void
 sync_listen(addr, port, sms)
         char *addr, *port;
@@ -780,14 +924,14 @@ sync_listen(addr, port, sms)
 	optval = 1;
 	if ((setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
 	    &optval, sizeof(optval))) != 0) {
-		syslog(LOG_ERR, "cannot set SO_REUSEADDR: %s",
+		mg_log(LOG_ERR, "cannot set SO_REUSEADDR: %s",
 		    strerror(errno));
 	}
 
 	optval = 1;
 	if ((setsockopt(s, SOL_SOCKET, SO_KEEPALIVE,
 	    &optval, sizeof(optval))) != 0) {
-		syslog(LOG_ERR, "cannot set SO_KEEPALIVE: %s",
+		mg_log(LOG_ERR, "cannot set SO_KEEPALIVE: %s",
 		    strerror(errno));
 	}
 
@@ -796,14 +940,14 @@ sync_listen(addr, port, sms)
 		optval = 1;
 		if ((setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY,
 		    &optval, sizeof(optval))) != 0) {
-			syslog(LOG_ERR, "cannot set IPV6_V6ONLY: %s",
+			mg_log(LOG_ERR, "cannot set IPV6_V6ONLY: %s",
 			    strerror(errno));
 		}
 	}
 #endif
 
 	if (bind(s, SA(&laddr), laddrlen) != 0) {
-		syslog(LOG_ERR, "cannot start MX sync, bind failed: %s",
+		mg_log(LOG_ERR, "cannot start MX sync, bind failed: %s",
 		    strerror(errno));
 		sms->runs = 0;
 		close(s);
@@ -811,7 +955,7 @@ sync_listen(addr, port, sms)
 	}
 
 	if (listen(s, MXGLSYNC_BACKLOG) != 0) {
-		syslog(LOG_ERR, "cannot start MX sync, listen failed: %s",
+		mg_log(LOG_ERR, "cannot start MX sync, listen failed: %s",
 		    strerror(errno));
 		sms->runs = 0;
 		close(s);
@@ -833,14 +977,21 @@ sync_server(arg)
 	char *addrstr;
 	char *from;
 	char *rcpt;
+	char from_clean[ADDRLEN + 1];
+	char rcpt_clean[ADDRLEN + 1];
 	char *datestr;
+	char *awstr;
 	char *cookie;
 	char line[LINELEN + 1];
 	peer_sync_t action;
 	sockaddr_t addr;
 	socklen_t addrlen;
 	time_t date;
+	time_t aw;
+	
+	conf_retain();
 
+	aw = time(NULL) + conf.c_autowhite_validity;
 	fprintf(stream, "200 Yeah, what do you want?\n");
 	fflush(stream);
 
@@ -856,7 +1007,7 @@ sync_server(arg)
 		fflush(stream);
 
 		/*
-		 * Get the command { quit | help | add | del }
+		 * Get the command { quit | help | add | del | del2 | flush }
 		 */
 		cookie = NULL;
 		if ((cmd = strtok_r(line, sep, &cookie)) == NULL) {
@@ -870,10 +1021,21 @@ sync_server(arg)
 		} else if ((strncmp(cmd, "help", CMDLEN)) == 0) {
 			sync_help(stream);
 			continue;
+		} else if ((strncmp(cmd, "vers3", CMDLEN)) == 0) {
+			sync_vers(stream, 3);
+			continue;
+		} else if ((strncmp(cmd, "vers2", CMDLEN)) == 0) {
+			sync_vers(stream, 2);
+			continue;
 		} else if ((strncmp(cmd, "add", CMDLEN)) == 0) {
 			action = PS_CREATE;
+		} else if ((strncmp(cmd, "del2", CMDLEN)) == 0) {
+			action = PS_DELETE;
+			aw = -1;
 		} else if ((strncmp(cmd, "del", CMDLEN)) == 0) {
 			action = PS_DELETE;
+		} else if ((strncmp(cmd, "flush", CMDLEN)) == 0) {
+			action = PS_FLUSH;
 		} else {
 			fprintf(stream, "102 Invalid command \"%s\"\n", cmd);
 			fflush(stream);
@@ -910,6 +1072,13 @@ sync_server(arg)
 			continue;
 		}	
 
+		if (action == PS_FLUSH) {
+			from = NULL;
+			rcpt = NULL;
+			date = 0;
+			goto eol;
+		}
+
 		/*
 		 * get { "from" email_address }
 		 */
@@ -931,6 +1100,8 @@ sync_server(arg)
 			fflush(stream);
 			continue;
 		}
+		(void)strncpy_rmsp(from_clean, from, ADDRLEN);
+		from = from_clean;
 
 		/*
 		 * get { "rcpt" email_address }
@@ -953,6 +1124,8 @@ sync_server(arg)
 			fflush(stream);
 			continue;
 		}
+		(void)strncpy_rmsp(rcpt_clean, rcpt, ADDRLEN);
+		rcpt = rcpt_clean;
 
 		/*
 		 * get { "date" valid_date }
@@ -978,9 +1151,36 @@ sync_server(arg)
 
 		date = atoi(datestr);
 
+		if (aw == -1) {
+			/*
+			 * get { "aw" valid_date }
+			 */
+			if ((keyword = strtok_r(NULL, sep, &cookie)) == NULL) {
+				fprintf(stream, "103 Incomplete command\n");
+				fflush(stream);
+				continue;
+			}
+
+			if (strncmp(keyword, "aw", CMDLEN) != 0) {
+				fprintf(stream, 
+				    "104 Unexpected keyword \"%s\"\n", keyword);
+				fflush(stream);
+				continue;
+			}
+
+			if ((awstr = strtok_r(NULL, sep, &cookie)) == NULL) {
+				fprintf(stream, "103 Incomplete command\n");
+				fflush(stream);
+				continue;
+			}
+
+			aw = atoi(awstr);
+		}
+
 		/* 
 		 * Check nothing remains
 		 */
+eol:
 		if ((keyword = strtok_r(NULL, sep, &cookie)) != NULL) {	
 			fprintf(stream, 
 			    "104 Unexpected keyword \"%s\"\n", keyword);
@@ -992,14 +1192,22 @@ sync_server(arg)
 		fflush(stream);
 
 		if (action == PS_CREATE) {
-			PENDING_WRLOCK;
-			pending_get(SA(&addr), addrlen, from, rcpt, date);
+			int dirty = 0;
+			PENDING_LOCK;
+			/* delay = -1 means unused: we supply the date */
+			if (pending_get(SA(&addr), addrlen, from, rcpt, date))
+				++dirty;
 			PENDING_UNLOCK;
+			dump_touch(dirty);
 		}
 		if (action == PS_DELETE) {
 			pending_del(SA(&addr), addrlen, from, rcpt, date);
-			autowhite_add(SA(&addr), addrlen, from, rcpt, NULL,
-			    "(mxsync)");
+			autowhite_add(SA(&addr), addrlen, from, 
+			    rcpt, &aw, "(mxsync)");
+		}
+		if (action == PS_FLUSH) {
+			pending_del_addr(SA(&addr), addrlen, NULL, 0);
+			autowhite_del_addr(SA(&addr), addrlen);
 		}
 
 		/* Flush modifications to disk */
@@ -1007,12 +1215,30 @@ sync_server(arg)
 	}
 
 	fprintf(stream, "202 Good bye\n");
-	fclose(stream);
+	Fclose(stream);
+
+	conf_release();
 
 	return;
 
 }
 
+static void
+sync_vers(stream, vers)
+	FILE *stream;
+	int vers;
+{
+	if (vers <= SYNC_PROTO_CURRENT) {
+		fprintf(stream, 
+		    "%d Yes, I speak version %d, what do you think?\n",
+		    800 + vers, vers);
+	} else {
+		fprintf(stream, "108 Invalid vers%d command\n", vers);
+	}
+	fflush(stream);
+	return;
+}
+
 void
 sync_help(stream)
 	FILE *stream;
@@ -1022,12 +1248,21 @@ sync_help(stream)
 	fprintf(stream, "203 Available commands are:\n");
 	fprintf(stream, "203 help  -- displays this message\n");
 	fprintf(stream, "203 quit  -- terminate connexion\n");
+	fprintf(stream, "203 vers2 -- speak version 2 protocol\n");
+	fprintf(stream, "203 vers3 -- speak version 3 protocol\n");
 	fprintf(stream, 
 	    "203 add addr <ip> from <email> rcpt <email> date <time>  "
 	    "-- add en entry\n");
 	fprintf(stream, 
 	    "203 del addr <ip> from <email> rcpt <email> date <time>  "
 	    "-- remove en entry\n");
+	fprintf(stream, 
+	    "203 del2 addr <ip> from <email> rcpt <email> date <time> "
+	    "aw <time> -- remove en entry, adding it to autowhite with "
+	    "given delay (version 2 only)\n");
+	fprintf(stream, 
+	    "203 flush addr <ip> -- remove anything about an ip "
+	    " (version 3 only)\n");
 	fflush(stream);
 
 	return;
@@ -1058,10 +1293,11 @@ sync_waitdata(fd)
 
 
 void
-sync_queue(peer, type, pending)	/* peer list must be read-locked */
+sync_queue(peer, type, pending, autowhite)/* peer list must be read-locked */
 	struct peer *peer;
 	peer_sync_t type;
 	struct pending *pending;
+	time_t autowhite;
 {
 	int error;
 	struct sync *sync;
@@ -1070,13 +1306,14 @@ sync_queue(peer, type, pending)	/* peer list must be read-locked */
 		return;
 
 	if ((sync = malloc(sizeof(*sync))) == NULL) {
-		syslog(LOG_ERR, "cannot allocate memory: %s", 
+		mg_log(LOG_ERR, "cannot allocate memory: %s", 
 		    strerror(errno)); 
 		exit(EX_OSERR);
 	}
 
 	sync->s_peer = peer;
 	sync->s_type = type;
+	sync->s_autowhite = autowhite;
 	sync->s_pending = pending_ref(pending);
 
 	/*
@@ -1084,13 +1321,16 @@ sync_queue(peer, type, pending)	/* peer list must be read-locked */
 	 * void it, but do not accept new entries anymore.
 	 */
 	if (!sync_queue_poke(peer, sync)) {
-		syslog(LOG_ERR, "peer %s queue overflow (%d entries), "
+		mg_log(LOG_ERR, "peer %s queue overflow (%d entries), "
 		    "discarding new entry", peer->p_name, peer->p_qlen);
 		sync_free(sync);
 	}
 
+	pthread_mutex_lock(&sync_dirty_lock);
+	sync_dirty = 1;
+	pthread_mutex_unlock(&sync_dirty_lock);
 	if ((error = pthread_cond_signal(&sync_sleepflag)) != 0) {
-		syslog(LOG_ERR, 
+		mg_log(LOG_ERR, 
 		    "cannot wakeup sync_sender: %s", strerror(error));
 		exit(EX_SOFTWARE);
 	}
@@ -1112,11 +1352,11 @@ sync_sender_start(void) {
 
 	if ((error = pthread_create(&tid, NULL, 
 	    (void *(*)(void *))sync_sender, NULL)) != 0) {
-		syslog(LOG_ERR, "pthread_create failed: %s", strerror(error));
+		mg_log(LOG_ERR, "pthread_create failed: %s", strerror(error));
 		exit(EX_OSERR);
 	}
 	if ((error = pthread_detach(tid)) != 0) {
-		syslog(LOG_ERR, "pthread_detach failed: %s", strerror(error));
+		mg_log(LOG_ERR, "pthread_detach failed: %s", strerror(error));
 		exit(EX_OSERR);
 	}
 	return;
@@ -1135,23 +1375,28 @@ sync_sender(dontcare)
 	int error;
 
 	if ((error = pthread_mutex_init(&mutex, NULL)) != 0) {
-		syslog(LOG_ERR, "pthread_mutex_init failed: %s", 
+		mg_log(LOG_ERR, "pthread_mutex_init failed: %s", 
 		    strerror(error));
 		exit(EX_OSERR);
 	}
 
 	if ((error = pthread_mutex_lock(&mutex)) != 0) {
-		syslog(LOG_ERR, "pthread_mutex_lock failed: %s", 
+		mg_log(LOG_ERR, "pthread_mutex_lock failed: %s", 
 		    strerror(error));
 		exit(EX_OSERR);
 	}
 
 	for (;;) {
-		if ((error = pthread_cond_wait(&sync_sleepflag, &mutex)) != 0)
-			syslog(LOG_ERR, "pthread_cond_wait failed: %s\n", 
-			    strerror(error));
+		pthread_mutex_lock(&sync_dirty_lock);
+		while (!sync_dirty)
+			pthread_cond_wait(&sync_sleepflag, &sync_dirty_lock);
+		sync_dirty = 0;
+		pthread_mutex_unlock(&sync_dirty_lock);
+
+		conf_retain();
+
 		if (conf.c_debug) {
-			syslog(LOG_DEBUG, "sync_sender running");
+			mg_log(LOG_DEBUG, "sync_sender running");
 			gettimeofday(&tv1, NULL);
 		}
 		done = 0;
@@ -1168,7 +1413,7 @@ sync_sender(dontcare)
 			while ((sync = sync_queue_peek(peer)) != NULL ) {
 
 				if (sync_send(sync->s_peer, sync->s_type, 
-				    sync->s_pending) != 0) {
+				    sync->s_pending, sync->s_autowhite) != 0) {
 					if (!sync_queue_poke(peer, sync))
 						sync_free(sync);
 
@@ -1186,10 +1431,12 @@ out:
 		if (conf.c_debug) {
 			gettimeofday(&tv2, NULL);
 			timersub(&tv2, &tv1, &tv3);
-			syslog(LOG_DEBUG, "sync_sender sleeping, "
+			mg_log(LOG_DEBUG, "sync_sender sleeping, "
 			    "done %d entries in %ld.%06lds", done,
 			    tv3.tv_sec, tv3.tv_usec);
 		}
+
+		conf_release();
 	}
 }
 
@@ -1215,25 +1462,29 @@ local_addr(sa, salen)
 #endif
 
 	default:
-		syslog(LOG_ERR, "local_addr: unsupported AF %d\n",
+		mg_log(LOG_ERR, "local_addr: unsupported AF %d",
 		    sa->sa_family);
 		return -1;
 		break;
 	}
 
 	if ((sfd = socket(sa->sa_family, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
-		syslog(LOG_ERR, "local_addr: socket failed: %s\n",
+		mg_log(LOG_ERR, "local_addr: socket failed: %s",
 		    strerror(errno));
 		return -1;
 	}
 
+	errno = 0;	 /* Solaris' bind() does not set errno... */
 	if (bind(sfd, sa, salen) == -1) {
 		if (errno != EADDRNOTAVAIL && 
 #ifdef __FreeBSD__
 		    errno != EINVAL &&
+#endif
+#ifdef __sun
+		    errno != 0 &&
 #endif
 		    1) {
-			syslog(LOG_ERR, "local_addr: bind failed: %s\n",
+			mg_log(LOG_ERR, "local_addr: bind failed: %s",
 			    strerror(errno));
 			islocal = -1;
 		} else {
@@ -1247,3 +1498,86 @@ local_addr(sa, salen)
 
 	return islocal;
 }
+
+static int 
+select_protocol(peer, s, stream)
+	struct peer *peer;
+	int s;
+	FILE *stream;
+{
+	char line[LINELEN + 1];
+	int vers;
+	char *replystr;
+	int replycode;
+	char sep[] = " \n\t\r";
+	char *cookie = NULL;
+
+	for (vers = SYNC_PROTO_CURRENT; vers > 1; vers--) {
+		fprintf(stream, "vers%d\n", vers);
+
+		fflush(stream);
+
+		sync_waitdata(s);	
+		if (fgets(line, LINELEN, stream) == NULL) {
+			mg_log(LOG_ERR, "Lost connexion with peer %s: "
+			    "%s (%d entries queued)", 
+			    peer->p_name, strerror(errno), peer->p_qlen);
+			return 0;
+		}
+
+		if ((replystr = strtok_r(line, sep, &cookie)) == NULL) {
+			mg_log(LOG_ERR, "Unexpected reply \"%s\" from peer %s "
+			    "closing connexion (%d entries queued)", 
+			    line, peer->p_name, peer->p_qlen);
+			return 0;
+		}
+
+		fflush(stream);
+
+		replycode = atoi(replystr);
+		if (replycode != 800 + vers) {
+			mg_log(LOG_DEBUG, 
+			    "peer %s answered code %d to command vers%d",
+			    peer->p_name, replycode, vers);
+		} else {
+			return vers;
+		}
+	}
+
+	return 1;
+}
+
+void
+peer_flush(pending)
+	struct pending *pending;
+{
+	struct peer *peer;
+
+	PEER_RDLOCK;
+	if (LIST_EMPTY(&peer_head))
+		goto out;
+
+	LIST_FOREACH(peer, &peer_head, p_list) {
+		/* Unsupported before verseion 3 */
+		if (peer->p_vers < 3)
+			continue;
+		sync_queue(peer, PS_FLUSH, pending, -1); /* -1: unused */
+	}
+
+out:
+	PEER_UNLOCK;
+	
+	return;
+}
+
+static int 
+is_fatal(err)
+	int err;
+{
+	int *i;
+
+	for (i = sync_ignored_errno; *i; i++)
+		if (errno == *i)
+			return 0;
+	return 1;
+}
