@@ -1,7 +1,8 @@
-/* $Id: sync.h,v 1.15 2004/11/12 14:22:56 manu Exp $ */
+/* $Id: sync.h,v 1.20 2007/03/16 03:58:04 manu Exp $ */
+/* vim: set sw=8 ts=8 sts=8 noet cino=(0: */
 
 /*
- * Copyright (c) 2004 Emmanuel Dreyfus
+ * Copyright (c) 2004-2007 Emmanuel Dreyfus
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -55,10 +56,6 @@ typedef unsigned int socklen_t;
 #define PEER_RDLOCK RDLOCK(peer_lock);
 #define PEER_UNLOCK UNLOCK(peer_lock);
 
-#define SYNC_WRLOCK WRLOCK(sync_lock);
-#define SYNC_RDLOCK RDLOCK(sync_lock);
-#define SYNC_UNLOCK UNLOCK(sync_lock);
-
 LIST_HEAD(peerlist, peer);
 TAILQ_HEAD(synclist, sync);
 
@@ -66,20 +63,25 @@ struct peer {
 	char *p_name;
 	FILE *p_stream;
 	int p_socket;
+	/* p_mtx protects p_deferred and p_qlen.
+	 * peer list must be read or rw locked before. */
+	pthread_mutex_t p_mtx;
 	struct synclist p_deferred;
 	LIST_ENTRY(peer) p_list;
-	size_t p_qlen;
+	unsigned int p_qlen;
 	int p_flags;
+	int p_vers;
 };
 
 #define P_LOCAL	1
 
-typedef enum { PS_CREATE, PS_DELETE } peer_sync_t;
+typedef enum { PS_CREATE, PS_DELETE, PS_DELETE2, PS_FLUSH } peer_sync_t;
 
 struct sync {
 	struct peer *s_peer;
 	struct pending *s_pending;
 	peer_sync_t s_type;
+	time_t s_autowhite;
 	TAILQ_ENTRY(sync) s_list;
 };
 
@@ -88,15 +90,17 @@ void peer_clear(void);
 void peer_add(char *);
 int peer_connect(struct peer *);
 void peer_create(struct pending *);
-void peer_delete(struct pending *);
+void peer_delete(struct pending *, time_t);
+void peer_flush(struct pending *);
 
-int sync_send(struct peer *, peer_sync_t,  struct pending *);
+int sync_send(struct peer *, peer_sync_t,  struct pending *, time_t);
 void sync_sender_start(void);
-void sync_queue(struct peer *, peer_sync_t, struct pending *);
+void sync_queue(struct peer *, peer_sync_t, struct pending *, time_t);
 void sync_free(struct sync *);
 
 void sync_sender(void *);
 void sync_master_restart(void);
+void sync_master_stop(void);
 void *sync_master(void *);
 void sync_server(void *);
 void sync_help(FILE *);
